""" Third step of the algorithm."""

import sys

class Triple:
    """
        a triple is (subject,predicate,object)
        for x in (subject,predicate,object), x is:
            either a string (= fixed information)
            or an int (= unknown)
    """
    def __init__(self, subjectT = None, predicateT = None, objectT = None ):
        self.subjectT = subjectT
        self.predicateT = predicateT
        self.objectT = objectT

    def renameUnknown(self,x,new_x):
        """
            unknown x is replaced by new_x everywhere it appears (means that x = x_new), x_new can also be a word (=replace an unknown by words)
        """
        if self.subjectT == x:
            self.subjectT = new_x
        if self.predicateT == x:
            self.predicateT = new_x
        if self.objectT == x:
            self.objectT = new_x

    def tripleUnit(self,t):
        """
            triple to string
        """
        if isinstance(t,int):
            if t == 0:
                return '??'
            else:
                return '?' + str(t)
        else:
            return t

    def __str__(self):
        return '(%s | %s | %s)' % (self.tripleUnit(self.subjectT),self.tripleUnit(self.predicateT),self.tripleUnit(self.objectT))

class TriplesBucket:
    """
        set of triples
    """
    def __init__(self, bucket = None):
        self.bucket = bucket or []

    def addTriple(self,t):
        self.bucket.append(t)

    def removeTriple(self,t):
        self.bucket.remove(t)
        
    def isEmpty(self): # true iff empty
        return not self.bucket
        
    def renameUnknown(self,x,new_x):
        for t in self.bucket:
            t.renameUnknown(x,new_x)

    def extractTriple(self,x): # extract a triple that contains the unknown x
        for t in self.bucket:
            if t.subjectT == x or t.predicateT == x or t.objectT == x:
                self.removeTriple(t)
                return t
        return None
    
    def __str__(self):
        return '\n'.join(str(x) for x in self.bucket)

#####################
# Triple production #
#####################

def genericProduce(t,nodeToID,rule,suffix=""):
    """
        Produce an element of a triple.
        The rule is a string in {'a','?A','c','?C'}
    """
    if suffix != '':
        suffix = ' %s' % suffix
    if rule == 'a':
        return '%s%s' % (t.parent.getWords(), suffix)
    if rule == '?A':
        return nodeToID[t.parent]
    if rule == 'c':
        return '%s%s' % (t.getWords(), suffix)
    if rule == '?C':
        return nodeToID[t]

def genericTripleProduce(t,nodeToID,rule,suffix=""):
    """
        Produce a triple, following the given rule.
        The rule is a 3-tuple of string, representing a triple.
        Example: rule=('?A','a','?C')
    """
    return Triple(genericProduce(t,nodeToID,rule[0]),
                    genericProduce(t,nodeToID,rule[1],suffix),
                    genericProduce(t,nodeToID,rule[2]))

# Rules of production of triples

def tripleProduce0(t,nodeToID,triplesBucket):
    pass

def tripleProduce1(t,nodeToID,triplesBucket):
    """
        a -b-> c : (c,a,?A) if c is a leaf
                   ?A = ?C otherwise
    """
    assert t.parent is not None
    if not t.child:
        triplesBucket.addTriple(genericTripleProduce(t,nodeToID,('c','a','?A')))
    else:
        triplesBucket.renameUnknown(nodeToID[t],nodeToID[t.parent])
        nodeToID[t] = nodeToID[t.parent]

def tripleProduce2(t,nodeToID,triplesBucket,suffix=''):
    """
        a -b-> c : (?A,a,c) if c is a leaf
                   (?A,a,?C) otherwise
        suffix: for prep_x
    """
    assert t.parent is not None
    if not t.child:
        triplesBucket.addTriple(genericTripleProduce(t,nodeToID,('?A','a','c'),suffix))
    else:
        triplesBucket.addTriple(genericTripleProduce(t,nodeToID,('?A','a','?C'),suffix))

def tripleProduce3(t,nodeToID,triplesBucket):
    """
        a -b-> c : (?A,c,a) if c is a leaf
                   (?A,?C,a) otherwise
    """
    assert t.parent is not None
    if not t.child:
        triplesBucket.addTriple(genericTripleProduce(t,nodeToID,('?A','c','a')))
    else:
        triplesBucket.addTriple(genericTripleProduce(t,nodeToID,('?A','?C','a')))

def tripleProduce4(t,nodeToID,triplesBucket):
    """
        a -b-> c : (c,a,?A) if c is a leaf
                   (?C,a,?A) otherwise
    """
    assert t.parent is not None
    if not t.child:
        triplesBucket.addTriple(genericTripleProduce(t,nodeToID,('c','a','?A')))
    else:
        triplesBucket.addTriple(genericTripleProduce(t,nodeToID,('?C','a','?A')))

def tripleProduce5(t,nodeToID,triplesBucket):
    """
        a -b-> c : (a,c,?A) if c is a leaf
                   (a,?C,?A) otherwise <-- or exit error? because ?C is strange
    """
    assert t.parent is not None
    if not t.child:
        triplesBucket.addTriple(genericTripleProduce(t,nodeToID,('a','c','?A')))
    else:
        triplesBucket.addTriple(genericTripleProduce(t,nodeToID,('a','?C','?A')))

def tripleProduce6(t,nodeToID,triplesBucket):
    """
        a -b-> c : ?A = c
    """
    assert t.parent is not None
    triplesBucket.renameUnknown(nodeToID[t.parent],t.getWords())


tripleMap = {
    't0'    : tripleProduce0,
    't1'    : tripleProduce1,
    't2'    : tripleProduce2,
    't3'    : tripleProduce3,
    't4'    : tripleProduce4,
    't5'    : tripleProduce5,
    't6'    : tripleProduce6
}

def initUnknowns(t,nodeToID,unknown=0):
    """
        assign a different unknown to each node
    """
    nodeToID[t] = unknown
    unknown += 1
    for c in t.child:
        unknown = initUnknowns(c,nodeToID,unknown)
    return unknown

def fillBucket(t,nodeToID,triplesBucket,tmap=tripleMap):
    """
        triple analysis of node t
    """
    if t.dependency in tmap:
        tmap[t.dependency](t,nodeToID,triplesBucket)
    if t.dependency.startswith('prep'): # prep_x or prepc_x
        prep = t.dependency[t.dependency.index('_')+1:] #_+ could be removed
        tripleProduce4(t,nodeToID,triplesBucket) #_+ instead of tripleProduce2(t,nodeToID,triplesBucket,prep) <--- preposition always remove now
    for c in t.child: # could become necessary to perform this step before
        fillBucket(c,nodeToID,triplesBucket)

questionMap = {
    # close question word
    'is'            : 'yesno',
    'are'           : 'yesno',
    'am'            : 'yesno',
    'was'           : 'yesno',
    'were'          : 'yesno',
    'will'          : 'yesno',
    'do'            : 'yesno',
    'does'          : 'yesno',
    'did'           : 'yesno',
    'have'          : 'yesno',
    'had'           : 'yesno',
    'has'           : 'yesno',
    'can'           : 'yesno',
    'could'         : 'yesno',
    'should'        : 'yesno',
    'shall'         : 'yesno',
    'may'           : 'yesno',
    'might'         : 'yesno',
    'would'         : 'yesno',
    # open question word
    'what'          : 'definition',
    'what kind'     : 'description',
    'what type'     : 'type',
    'what sort'     : 'type',
    'what time'     : 'time',
    'when'          : 'time',
    'why'           : 'reason',
    'where'         : 'place',
    'who'           : 'person',
    'how'           : 'manner',
    'how much'      : 'amount',
    'how many'      : 'quantity',
    'how old'       : 'age',
    'how far'       : 'distance',
    'how long'      : 'length',
    'how tall'      : 'height',
    'how deep'      : 'depth',
    'how wide'      : 'width',
    'how fast'      : 'speed',
    'how often'     : 'frequency',
    'how come'      : 'reason',
    'which'         : 'choice',
    'whom'          : 'person',
    'whose'         : 'possession'
        # how big
}

def buildBucket(t,qw):
    """
        return a TriplesBucket associated to tree t and question word qw
    """
    nodeToID = {}
    triplesBucket = TriplesBucket()
    initUnknowns(t,nodeToID)
    #_+ triplesBucket.addTriple(Triple(nodeToID[t.child[0]],questionMap[qw],nodeToID[t])) # process the question word
    tripleProduce1(t.child[0],nodeToID,triplesBucket) #_+ instead of question word
    fillBucket(t,nodeToID,triplesBucket)
    return triplesBucket



import sys

#########################
# Quotation recognition #
#########################

def findQuotations(r):
    """
        Return a list of elements of the form (begin,end,set of integers).
        Each set is a set of words index belonging to a same quotation.
        Begin and end are the index of the quotations marks
    """
    index=0
    inQuote=False
    quotationList=[]
    quotationSet = set()
    for word in r['words']:
        index+=1
        if word[0] == "``":
            assert not inQuote
            inQuote = True
            begin=index
            continue
        if word[0] == "''":
            assert inQuote
            inQuote=False
            quotationList+=[(begin,index,quotationSet)]
            quotationSet = set()
            continue
        if inQuote:
            quotationSet.add(index)
    assert not inQuote
    return quotationList

def matchingQuoteWord(w,quotationList):
    """
        Return the quotation set in which belong the word (None otherwise).
        w must be of the form (word,index)
    """
    for quote in quotationList:
        if w[1] in quote[2]:
            return quote
    return None

def matchingQuote(wlist,quotationList):
    """
        Return the quotation set in which belong all the words of the list (None otherwise).
        If two words does not belong to the same quotation, error.
    """
    quote=matchingQuoteWord(wlist[0],quotationList)
    for w in wlist:
        if matchingQuoteWord(w,quotationList) != quote:
            sys.exit('exit: node belong to several quotations (please, report your sentence on http://goo.gl/EkgO5l)\n%s\n' % self.getWords())
    return quote

def quotationTraversal(t,quotationList,quoteIndexToNode):
    """
        Traverse the tree to merge quotations, given a quotationList (computed
            with findQuotation).
        Fill quoteIndexToNode (map from the index of the beginning of the quote to the node.
    """
    childCopy = list(t.child)
    for c in childCopy:
        quotationTraversal(c,quotationList,quoteIndexToNode)
    quote = matchingQuote(t.wordList,quotationList)
    if not quote:
        return
    if not quote[0] in quoteIndexToNode: 
        quoteIndexToNode[quote[0]] = t
    childCopy = list(t.child)
    for c in childCopy:
        if matchingQuote(c.wordList,quotationList) == quote:
            t.merge(c,True)
            quoteIndexToNode[quote[0]] = t

def handleLostQuotationWords(r,quoteIndexToNode):
    """
        Add quotation words deleted by Stanford CoreNLP, such as "in" or "of".
    """
    inQuote=False
    quoteNode=None
    index=0
    # Add words which are not nodes
    for word in r['words']:
        index+=1
        if word[0] == "``":
            inQuote = True
            quoteNode=quoteIndexToNode[index]
            continue
        if word[0] == "''":
            inQuote = False
            quoteNode.wordList.sort(key = lambda x: x[1])
            continue
        if inQuote and (word[0],index) not in quoteNode.wordList:
            quoteNode.wordList += [(word[0],index)]

def addQuotationTag(quoteIndexToNode):
    """
        Add the tag "QUOTATION" to all nodes in quoteIndexToNode.
    """
    for c in quoteIndexToNode.values():
        c.namedEntityTag="QUOTATION"

def mergeQuotations(t,r):
    """
        Merge all nodes corresponding to quotations.
    """
    quotationList = findQuotations(r)
    quoteIndexToNode = {}
    # Merge existing nodes belonging to quotations.
    quotationTraversal(t,quotationList,quoteIndexToNode)
    handleLostQuotationWords(r,quoteIndexToNode)
    addQuotationTag(quoteIndexToNode)

###################
# NER recognition #
###################

def mergeNamedEntityTagChildParent(t):
    """
        Merge all nodes n1,n2 such that:
            * n1 is parent of n2
            * n1 and n2 have a same namedEntityTag
    """
    for c in t.child:
        mergeNamedEntityTagChildParent(c)
    sameTagChild = set()
    if t.namedEntityTag == 'undef':
        return
    for c in t.child:
        if c.namedEntityTag == t.namedEntityTag:
            sameTagChild.add(c)
    for c in sameTagChild:
        t.merge(c,True)

def mergeNamedEntityTagSisterBrother(t):
    """
        Merge all nodes n1,n2 such that:
            * n1 and n2 have a same parent
            * n1 and n2 have a same namedEntityTag
            * n1 and n2 have a same dependency
    """
    for c in t.child:
        mergeNamedEntityTagSisterBrother(c)
    tagToNodes = {}
    for c in t.child:
        if c.namedEntityTag == 'undef':
            continue
        try:
            tagToNodes[c.namedEntityTag+c.dependency].add(c)
        except KeyError:
            tagToNodes[c.namedEntityTag+c.dependency] = set([c])
    for sameTag in tagToNodes.values():
        x = sameTag.pop()
        for other in sameTag:
            x.merge(other,True)

def mergeNamedEntityTag(t):
    mergeNamedEntityTagChildParent(t)
    mergeNamedEntityTagSisterBrother(t)



"""Configuration module."""

import os
import json
import logging
from ppp_core.exceptions import InvalidConfig


class Config:
    __slots__ = ('corenlp_server')
    def __init__(self, data=None):
        if not data:
            try:
                with open(self.get_config_path()) as fd:
                    data = json.load(fd)
            except ValueError as exc:
                raise InvalidConfig(*exc.args)
        self.corenlp_server = data['corenlp_server']

    @staticmethod
    def get_config_path():
        path = os.environ.get('PPP_NLP_CLASSICAL_CONFIG', '')
        if not path:
            raise InvalidConfig('Could not find config file, please set '
                                'environment variable $PPP_NLP_CLASSICAL_CONFIG.')
        return path



import sys

"""
    Taken from: http://www.interopia.com/education/all-question-words-in-english/
    Yes/no question
"""
closeQuestionWord = [
    'is', 'are', 'am', 'was', 'were', 'will', 'do', 'does', 'did', 'have', 'had', 'has', 'can', 'could', 'should', 'shall', 'may', 'might', 'would'
]

"""
    Open-ended questions 
    + What... for, What... like, Why don't, Where from?
    Rarely used: Wherefore, Whatever, Wherewith, Whither, Whence, However
"""
openQuestionWord = [
    'what', 'what kind', 'what type', 'what sort', 'what time', 'when', 'why', 'where', 'who', 'how', 'how much', 'how many', 'how old', 'how far', 'how long', 'how tall', 'how deep', 'how wide', 'how fast', 'how often', 'how come', 'which', 'whom', 'whose'
]

def removeWord(t,word):
    """
        Remove word (of type str*int = word*position_in_sentence) in t
        Assume word has no child
    """
    if word in t.wordList:
        if not t.child:
            t.parent.child.remove(t) 
        else:
            sys.exit('exit: question word has child (please, report your sentence on http://goo.gl/EkgO5l)\n')
    else:
        for c in t.child:
            removeWord(c,word)

def firstWords(t,start):
    """
        Put the 2 first words of the sentence in start (list of size 2)
    """
    for n in t.wordList:
        if n[1] == 1:
            start[0] = n
        elif n[1] == 2:
            start[1] =n
    for c in t.child:
        firstWords(c,start)

def identifyQuestionWord(t):
    """
        Identify, remove (if open qw) and return the question word
    """
    start = [None,None]
    firstWords(t,start)
    if not start[0]:
        sys.exit('exit: i don\'t understand (please, report your sentence on http://goo.gl/EkgO5l)')
    if start[1] and start[0][0].lower() + ' ' + start[1][0].lower() in openQuestionWord:
        removeWord(t,start[0])
        removeWord(t,start[1])
        return start[0][0].lower() + ' ' + start[1][0].lower()
    if start[0][0].lower() in openQuestionWord: 
        removeWord(t,start[0])
        return start[0][0].lower()
    if start[0][0].lower() in closeQuestionWord: 
        return start[0][0].lower()
    sys.exit('exit: question word not found (please, report your sentence on http://goo.gl/EkgO5l)')



"""Natural language processing module for the PPP."""

from ppp_core import HttpRequestHandler
from .preprocessingMerge import mergeNamedEntityTagChildParent, mergeNamedEntityTagSisterBrother, mergeNamedEntityTag
from .preprocessing import DependenciesTree, computeTree
from .questionIdentify import identifyQuestionWord
from .dependencyAnalysis import simplify
from .tripleProduction import buildBucket, Triple, TriplesBucket, tripleProduce1, tripleProduce2, tripleProduce3, tripleProduce4, tripleProduce5, tripleProduce6
from .treeTranslation import buildTree

#_+ from .requesthandler import RequestHandler

def app(environ, start_response):
    """Function called by the WSGI server."""
    return HttpRequestHandler(environ, start_response, RequestHandler) \
            .dispatch()

__all__ = ['DependenciesTree','computeTree','mergeNamedEntityTagChildParent','mergeNamedEntityTagSisterBrother','mergeNamedEntityTag','simplify', 'identifyQuestionWord','buildBucket','Triple','TriplesBucket', 'tripleProduce1', 'tripleProduce2', 'tripleProduce3','tripleProduce4','tripleProduce5','tripleProduce6','buildTree']




import sys
import ppp_datamodel
from .tripleProduction import TriplesBucket

def newSubNode(triplesBucket,t):
    """
        built the subtree of root t
    """
    if isinstance(t,int):
        sub_t = triplesBucket.extractTriple(t)
        return newNode(triplesBucket,sub_t)
    else:
        return ppp_datamodel.Resource(value=t)

def newNode(triplesBucket,t):
    """
        build a tree of root t, with all the triples of triplesBucket possibles
    """
    if not t:
        return ppp_datamodel.Missing()
    subj=newSubNode(triplesBucket,t.subjectT)
    pred=newSubNode(triplesBucket,t.predicateT)
    obj=newSubNode(triplesBucket,t.objectT)
    return ppp_datamodel.Triple(subject=subj, predicate=pred, object=obj)

def buildTree(triplesBucket):
    """
        build a tree from a triplesBucket
    """
    t = triplesBucket.extractTriple(0) # 0 = main unknown
    assert t # != None
    tree = newNode(triplesBucket,t)
    if not triplesBucket.isEmpty():
        sys.exit('exit: tree translation not completed (please, report your sentence on http://goo.gl/EkgO5l)\n')
    return tree



"""Request handler of the module."""

import json
import jsonrpclib

from ppp_datamodel import Sentence
from ppp_datamodel.communication import TraceItem, Response
from ppp_core.exceptions import ClientError

from .config import Config
from . import computeTree, simplify, buildTree, buildBucket

class StanfordNLP:
    def __init__(self, url):
        self.server = jsonrpclib.Server(url)

    def parse(self, text):
        return json.loads(self.server.parse(text))
stanfordnlp = StanfordNLP(Config().corenlp_server)

class RequestHandler:
    def __init__(self, request):
        self.request = request

    def answer(self):
        if not isinstance(self.request.tree, Sentence):
            return []
        result = stanfordnlp.parse(self.request.tree.value)
        tree = computeTree(result['sentences'][0])
        qw = simplify(tree)
        t = buildTree(buildBucket(tree,qw))
        m = {'accuracy': 0.5, 'relevance': 0.5}
        response = Response('en', t, m,
                            [TraceItem('NLP-Classical', t, m)])
        print(repr(response))
        return [response]



#!/usr/bin/env python3

from setuptools import setup, find_packages

setup(
    name='ppp_nlp_classical',
    version='0.2.1',
    description='Natural language processing module for the PPP.',
    url='https://github.com/ProjetPP/PPP-NLP-Classical',
    author='Projet Pensées Profondes',
    author_email='ppp2014@listes.ens-lyon.fr',
    license='MIT',
    classifiers=[
        'Environment :: No Input/Output (Daemon)',
        'Development Status :: 1 - Planning',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Software Development :: Libraries',
    ],
    install_requires=[
        'ppp_datamodel>=0.5,<0.6',
        'ppp_core>=0.5,<0.6',
        'jsonrpclib-pelix',
        'nltk'
    ],
    packages=[
        'ppp_nlp_classical',
    ],
)

import sys
if 'install' in sys.argv:
    import nltk
    nltk.download("wordnet")



""" Second step of the algorithm."""

import sys
from .questionIdentify import identifyQuestionWord

def remove(t):
    t.parent.child.remove(t)

def impossible(t):
    sys.exit('exit: %s dependency unexpected (please, report your sentence on http://goo.gl/EkgO5l)\n' % t)
    #remove(t)

def ignore(t):
    remove(t)

def merge(t):
    t.parent.merge(t,True)

dependenciesMap = {
    'undef'     : 't0', # personnal tag, should not happen?
    'root'      : 't0',
    'dep'       : 't6', # ? instead of t1
        'aux'       : remove,
            'auxpass'   : remove,
            'cop'       : impossible,
        'arg'       : impossible,
            'agent'     : 't4',
            'comp'      : 't2',
                'acomp'     : 't2',
                'ccomp'     : 't2',
                'xcomp'     : 't2',
                'pcomp'     : 't2', # -
                'obj'       : impossible,
                    'dobj'      : 't4', #_+ instead of t2
                    'iobj'      : 't2',
                    'pobj'      : 't2', # -
            'subj'      : impossible,
                'nsubj'     : 't1',
                    'nsubjpass'    : 't4', #_+ ? instead of  t3
                'csubj'     : impossible,
                    'csubjpass'    : impossible,
        'cc'        : impossible,
        'conj'      : 't0',
            'conj_and'  : ignore,
            'conj_or'   : ignore,
            'conj_negcc': ignore, #?
        'expl'      : ignore,
        'mod'       : 't3',
            'amod'      : 't5', #
            'appos'     : 't3',
            'advcl'     : 't3',
            'det'       : remove,
            'predet'    : ignore,
            'preconj'   : ignore,
            'vmod'      : 't2',
            'mwe'       : merge,
                'mark'      : ignore,
            'advmod'    : merge,
                'neg'       : 't0', # need a NOT node
            'rcmod'     : ignore,
                'quantmod'  : ignore,
            'nn'        : merge,
            'npadvmod'  : merge,
                'tmod'      : 't2',
            'num'       : merge,
            'number'    : merge,
            'prep'      : 't4', # ?
            'prepc'     : 't4', # ?
            'poss'      : 't4',
            'possessive': impossible,
            'prt'       : merge,
        'parataxis' : ignore, #  ?
        'punct'     : impossible,
        'ref'       : impossible,
        'sdep'      : impossible,
            'xsubj'     : 't2',
        'goeswith'  : merge,
        'discourse' : remove
}

def collapseDependency(t,depMap=dependenciesMap):
    """
        Apply the rules of depMap to t
    """
    temp = list(t.child) # copy, t.child is changed while iterating
    for c in temp:
        collapseDependency(c,depMap)
    try:
        if isinstance(depMap[t.dependency], str):
            t.dependency = depMap[t.dependency]
        else:
            depMap[t.dependency](t)
    except KeyError: # prep_x, prepc_x,... (others?) see the manual
        if (t.dependency[:t.dependency.index('_')] not in {'prep','prepc'}):
            sys.exit('exit: dependency unknown (please, report your sentence on http://goo.gl/EkgO5l)\n')
        pass


def simplify(t):
    """
            identify and remove question word
            collapse dependencies of tree t
    """
    s = identifyQuestionWord(t) # identify and remove question word
    collapseDependency(t) # apply dependency rules of collapsing
    return s



""" First step of the algorithm."""

import sys
from .preprocessingMerge import mergeQuotations, mergeNamedEntityTag
from nltk.stem.wordnet import WordNetLemmatizer

class DependenciesTree:
    """
        One node of the parse tree.
        It is a group of words of the initial sentence.
    """
    def __init__(self, word, namedentitytag='undef', dependency='undef', child=None, parent=None):
        self.wordList = [(word[:word.rindex('-')],int(word[word.rindex('-')+1:]))] # words of the node
        self.namedEntityTag = namedentitytag 
        self.dependency = dependency # dependency from self to its parent
        self.child = child or [] # children of self
        self.text = "" # each node contains whole sentence
        self.parent=parent # parent of self

    def string(self):
        # Concatenation of the words of the root
        w = self.getWords()
        s=''
        # Adding the definition of the root (dot format)
        t=''
        if(self.namedEntityTag != 'O' and self.namedEntityTag != 'undef'):
            t+= " [{0}]".format(self.namedEntityTag)
        s+="\t\"{0}\"[label=\"{1}{2}\",shape=box];\n".format(self.wordList[0][0]+str(self.wordList[0][1]),w,t)
        # Adding definitions of the edges
        for n in self.child:
            s+="\t\"{0}\" -> \"{1}\"[label=\"{2}\"];\n".format(self.wordList[0][0]+str(self.wordList[0][1]),n.wordList[0][0]+str(n.wordList[0][1]),n.dependency)
        # Recursive calls
        for n in self.child:
            s+=n.string()
        return s

    def __str__(self):
        """
            Print dependency graph in dot format
        """
        return "digraph relations {"+"\n{0}\tlabelloc=\"t\"\tlabel=\"{1}\";\n".format(self.string(),self.text)+"}"

    def merge(self,other,mergeWords):
        """
            Merge the root of the two given trees into one single node.
            Merge the two wordList if mergeWords=True (otherwise only keep the WordList of self).
            The result is stored in node 'self'.
        """
        self.child += other.child
        for c in other.child:
            c.parent = self
        if mergeWords:
          self.wordList = other.wordList + self.wordList
          self.wordList.sort(key = lambda x: x[1])
        if other.parent:
            other.parent.child.remove(other)
        other.wordList = [("merged",0)]

    def getWords(self):
        """
            concatenate all strings of the node (in wordList)
        """
        self.wordList.sort(key = lambda x: x[1]) 
        return ' '.join(x[0] for x in self.wordList)

def computeEdges(r,nameToNodes):
    """
        Compute the edges of the dependence tree.
        Take in input a piece of the result produced by StanfordNLP, and the
        map from names to nodes.
    """
    for edge in r['indexeddependencies']:
        try:
            n1 = nameToNodes[edge[1]]
        except KeyError:
            n1 = DependenciesTree(edge[1])
            nameToNodes[edge[1]] = n1
        try:
            n2 = nameToNodes[edge[2]]
        except KeyError:
            n2 = DependenciesTree(edge[2])
            nameToNodes[edge[2]] = n2
        # n1 is the parent of n2
        n1.child = n1.child+[n2]
        n2.parent = n1
        n2.dependency = edge[0]

def computeTags(r,nameToNodes):
    """
        Compute the tags of the dependence tree nodes.
        Take in input a piece of the result produced by StanfordNLP, and the
        map from names to nodes.
    """
    index=1
    # Computation of the tags of the nodes
    for word in r['words']:
        if word[0].isalnum() or word[0] == '$' or  word[0] == '%':
            w=word[0]+'-'+str(index) # key in the nameToNodes map
            index+=1
            try:
                n = nameToNodes[w]
                if word[1]['NamedEntityTag'] != 'O':
                    n.namedEntityTag = word[1]['NamedEntityTag']
            except KeyError:        # this node does not exists (e.g. 'of' preposition)
                pass

def initText(t,s):
    """
        Set text attribute for all nodes, with string s.
    """
    t.text = s
    for c in t.child:
        initText(c,s)

def normalizeWord(word,lmtzr):
    """
        Apply lemmatization to the given word.
    """
    result=lmtzr.lemmatize(word,'n')
    if len(result)<len(word):
        return result
    return lmtzr.lemmatize(word,'v')

def normalize(t,lmtzr):
    for c in t.child:
        normalize(c,lmtzr)
    if t.namedEntityTag == 'undef':
        t.wordList = list((normalizeWord(w[0],lmtzr),w[1]) for w in t.wordList)

def computeTree(r):
    """
        Compute the dependence tree.
        Take in input a piece of the result produced by StanfordNLP (if foo is this result, then r = foo['sentences'][0])
        Apply quotation and NER merging
        Return the root of the tree (word 'ROOT-0').
    """
    nameToNodes = {} # map from the original string to the node
    computeEdges(r,nameToNodes)
    computeTags(r,nameToNodes)
    tree = nameToNodes['ROOT-0']
    initText(tree,r['text'].replace('"','\\\"'))
    mergeQuotations(tree,r) # quotation merging
    mergeNamedEntityTag(tree) # NER merging
    lmtzr = WordNetLemmatizer()
    normalize(tree,lmtzr)
    return tree
